<?php

require_once 'libs/interfaces/interface-verifone-hosted-arrayable.php';
require_once 'libs/interfaces/interface-verifone-hosted-jsonable.php';
require_once 'libs/abstract/abstract-verifone-api-entity.php';
require_once 'libs/includes/class-verifone-hosted-logger.php';
require_once 'libs/includes/class-verifone-hosted-checkout.php';
require_once 'libs/includes/class-verifone-hosted-helper.php';
require_once 'libs/includes/class-verifone-hosted-customer.php';
require_once 'libs/includes/class-verifone-hosted-helper.php';
require_once 'libs/includes/class-verifone-refund.php';
require_once 'libs/VerifoneEcomAPI/autoload.php';

use VerifoneEcomAPI\ApiWrapper\Authentication\BasicAuth;
use VerifoneEcomAPI\ApiWrapper\Http\ApiClient\Client;
use VerifoneEcomAPI\ApiWrapper\Http\SimpleCurl;
use VerifoneEcomAPI\ApiWrapper\Regions\Emea;
use VerifoneEcomAPI\ApiWrapper\Regions\Na;
use VerifoneEcomAPI\ApiWrapper\Regions\NzAu;
use VerifoneEcomAPI\ApiWrapper\Settings;


use PrestaShop\PrestaShop\Core\Payment\PaymentOption;

if ( ! defined('_PS_VERSION_')) {
    exit;
}

/**
 * Class verifone - safe payment method
 */
class Verifone extends PaymentModule
{
    const DEBUG_MODE = false;
    public $details;
    public $owner;
    public $name;
    public $address;
    public $bootstrap;
    public $is_eu_compatible;
    public $extra_mail_vars;
    public $confirmUninstall;
    protected $_html = '';
    protected $requiredFields = [
        'VERIFONE_API_KEY',
        'VERIFONE_USER_ID',
        'VERIFONE_ENTITY_ID',
        'VERIFONE_TITLE',
    ];

    protected $allFields = [
        'VERIFONE_MODE',
        'VERIFONE_API_KEY',
        'VERIFONE_USER_ID',
        'VERIFONE_ENTITY_ID',
        'VERIFONE_REGION',
        'VERIFONE_CHECKOUT_TYPE',
        'VERIFONE_TITLE',
        'VERIFONE_DESCRIPTION',
        'VERIFONE_THEME_ID',
        'VERIFONE_CARD_ENABLED',
        'VERIFONE_PAYMENT_CONTRACT_ID',
        'VERIFONE_3DS_ENABLED',
        'VERIFONE_3DS_CONTRACT_ID',
        'VERIFONE_PAYPAL_ENABLED',
        'VERIFONE_PAYPAL_CONTRACT_ID',
        'VERIFONE_APPLE_PAY_ENABLED',
        'VERIFONE_APPLE_PAY_CONTRACT_ID',
        'VERIFONE_APPLE_PAY_THREE_DS_CONTRACT_ID',
        'VERIFONE_APPLE_PAY_SCA_LEVEL',
        'VERIFONE_GOOGLE_PAY_ENABLED',
        'VERIFONE_GOOGLE_PAY_CONTRACT_ID',
        'VERIFONE_GOOGLE_PAY_THREE_DS_CONTRACT_ID',
        'VERIFONE_GOOGLE_PAY_SCA_LEVEL',
        'VERIFONE_CURL_VERIFY_SSL',
        'VERIFONE_TRANSACTION_TYPE',
    ];

    /**
     * @var Client
     */
    public $client;

    /**
     * @var Verifone_Helper
     */
    public $verifone_helper;

    /**
     * @var Settings
     */
    public $verifone_settings;

    /**
     * @var Verifone_Cart_Logger
     */
    public $logger;

    /**
     * @var Emea|Na|NzAu
     */
    public $mapper;

    /**
     * @var SimpleCurl
     */
    public $simple_curl;

    /**
     * @var BasicAuth
     */
    public $auth;

    /**
     * verifone constructor.
     */
    public function __construct()
    {
        $settings                     = new Settings();
        $this->name                   = 'verifone';
        $this->tab                    = 'payments_gateways';
        $this->version                = '3.1.0';
        $this->ps_versions_compliancy = ['min' => '1.7', 'max' => _PS_VERSION_];
        $this->author                 = 'Verifone';
        $this->controllers            = ['validation'];
        $this->is_eu_compatible       = true;
        $this->currencies             = true;
        $this->currencies_mode        = 'checkbox';
        $this->verifone_settings      = $settings;
        $this->logger                 = new Verifone_Cart_Logger();
        $this->simple_curl            = new SimpleCurl();

        // mini factory :D
        switch ($this->verifone_settings->getRegion()) {
            case 'NZAU':
                $this->mapper = new NzAu($this->verifone_settings);
                break;
            case 'NA':
                $this->mapper = new Na($this->verifone_settings);
                break;
            default:
                $this->mapper = new Emea($this->verifone_settings);
                break;
        }

        $this->auth   = new BasicAuth($this->verifone_settings);
        $this->client = new Client($this->mapper, $this->verifone_settings, $this->auth, $this->simple_curl);

        $this->verifone_helper = new Verifone_Helper();
        $this->bootstrap       = true;
        parent::__construct();

        $this->displayName = $this->l('Verifone');
        $this->description = $this->l('Verifone safe payment solutions');

        if ( ! count(Currency::checkPaymentCurrencies($this->id))) {
            $this->warning = $this->l('No currency has been set for this module.');
        }
        $this->confirmUninstall = $this->l('Are you sure you want to uninstall Verifone payment modules?');
    }

    /**
     * install the module
     * @return bool
     */
    public function install()
    : bool
    {

        if (parent::install()
            && $this->registerHook('displayHeader')
            && $this->registerHook('paymentOptions')
            && $this->registerHook('displayOrderConfirmation')
            && $this->registerHook('actionProductCancel')
            && $this->registerHook('actionOrderStatusUpdate')
            && $this->registerHook('displayAdminOrder')
            && $this->registerHook('displayAdminOrderTop')
        ) {
            return true;
        }

        return false;
    }

    /**
     * uninstall the module and deletes the config keys
     * @return bool
     * @throws PrestaShopException
     */
    function uninstall()
    : bool
    {

        $configs = Configuration::getMultiple($this->allFields);
        foreach ($configs as $key => $config) {
            Configuration::deleteByName($key);

        }

        return $this->unregisterHook('displayHeader')
               && $this->unregisterHook('paymentOptions')
               && $this->unregisterHook('displayOrderConfirmation')
               && $this->unregisterHook('actionProductCancel')
               && $this->unregisterHook('actionOrderStatusUpdate')
               && $this->unregisterHook('displayAdminOrder')
               && $this->unregisterHook('displayAdminOrderTop')
               && parent::uninstall();
    }

    /**
     * show the settings page, also saves and validates the form on submit
     * @return string
     * @throws SmartyException
     * @throws PrestaShopException
     */
    public function getContent()
    : string
    {
        $output = '';
        // default/saved data
        $webhookUrl = $this->context->link->getModuleLink($this->name, 'webhook', [], true);
        $token      = Tools::getAdminTokenLite('AdminModules');
        $action     = AdminController::$currentIndex.'&configure='.$this->name.'&token='.$token;
        $this->context->smarty->assign('data', Configuration::getMultiple($this->allFields));
        $this->context->smarty->assign('action', $action);
        $this->context->smarty->assign('webhookUrl', $webhookUrl);
        if (Tools::isSubmit('submit'.$this->name)) {
            $this->reset_token();
            $postData      = Tools::getAllValues();
            $processedData = [];
            foreach ($postData as $key => $value) {
                if (strpos($key, 'VERIFONE_') !== false) {
                    if (strpos($key, '_ID') !== false) { // must be a valid UUID
                        $this->verifone_helper->uuid($key, $value);
                    }
                    if (in_array($key, $this->requiredFields)) { // mandatory fields
                        $this->verifone_helper->required($key, $value);
                    }
                    $processedData[$key] = trim($value);
                }
            }
            //save new data if ok
            if ($this->verifone_helper->hasErrors() === false) {
                // remove old data
                $configs = Configuration::getMultiple($this->allFields);
                foreach ($configs as $key => $config) {
                    Configuration::deleteByName($key);
                }
                foreach ($this->allFields as $key) {
                    Configuration::updateValue($key, $postData[$key] ?? null);
                }
                $output .= $this->displayConfirmation($this->l('Settings updated!'));
            } else {
                $this->context->smarty->assign('vf_errors', $this->verifone_helper->getErrorBag());
                $output .= $this->displayError($this->l('Please fix all the errors before you save!'));
            }
            //overwrite old data with the new data
            $this->context->smarty->assign('data', $postData);
        }

        $output .= $this->context->smarty->fetch('module:verifone/views/templates/back/admin_tabs.tpl');

        return $output;
    }


    /**
     * ==============================HOOKS==============================
     * @throws PrestaShopDatabaseException|SmartyException
     */

    public function hookHeader()
    : ?string
    {
        if (Tools::getValue('controller') == "order" && $this->checkActiveModule()) {
            Media::addJsDef(['verifone_verify_url' => $this->context->link->getModuleLink($this->name, 'generateUrl', ['module_name' => $this->name], true)]);
            $this->context->controller->registerJavascript($this->name.'-gen-url-context', 'modules/'.$this->name.'/views/assets/js/validate.js');
            $this->context->controller->registerJavascript($this->name.'-gen-url-context-iframe', 'modules/'.$this->name.'/views/assets/js/iframe.js');
        }

        return null;
    }

    /**
     * @param $params
     *
     * @return null[]|PaymentOption[]|void
     */
    public function hookPaymentOptions($params)
    {
        if ( ! $this->active || ! $this->checkCurrency($params['cart'])) {
            return;
        }
        // we clear the cache for every change we make
        Tools::clearSmartyCache();
        Tools::clearXMLCache();
        Media::clearCache();
        Tools::generateIndex();

        return [$this->getVerifonePaymentOption()];
    }

    /**
     * @param $params
     *
     * @return false|string
     */
    public function hookDisplayOrderConfirmation($params)
    {
        if (( ! isset($params['order']) || $params['order']->module != $this->name) || ! $this->active) {
            return false;
        }

        $order = $params['order'] ?? null;
        $cart  = $params['cart'] ?? null;
        $this->smarty->assign(['order' => $order, 'cart' => $cart]);

        return $this->fetch('module:verifone/views/templates/hook/payment_return.tpl');
    }

    /**
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException|SmartyException
     */
    public function hookDisplayAdminOrderTop($params)
    {

        $id_order = $params['id_order'] ?? null;
        $order    = new Order((int) $id_order);
        if (Validate::isLoadedObject($order) && $order->module === $this->name) {
            $errors = $this->getAdminOrderPageRefundErrors();

            if ( ! empty($errors)) {
                $this->context->controller->errors[] = $errors;
            }

            $return = $this->disablePartialRefund();
            $return .= $this->getRefundCommentBox();
            $return .= $this->getTotalItemsForRefund($order);

            return $return;
        }

        return false;
    }

    /**
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException|SmartyException
     */
    public function hookDisplayAdminOrder($params)
    {
        if (version_compare(_PS_VERSION_, '1.7.7', '>=')) {
            return false;
        }
        $id_order = $params['id_order'] ?? null;
        $order    = new Order((int) $id_order);
        if (Validate::isLoadedObject($order) && $order->module === $this->name) {
            $errors = $this->getAdminOrderPageRefundErrors();
            if ( ! empty($errors)) {
                $this->context->controller->errors[] = $errors;
            }
            $return = $this->disablePartialRefund();
            $return .= $this->getRefundCommentBox();
            $return .= $this->getTotalItemsForRefund($order);

            return $return;
        }

        return false;
    }


    /** REFUND
     *
     * @param $params
     */
    public function hookActionProductCancel($params)
    {
        $order = $params['order'] ?? null;
        $this->verifoneRefund($order);
    }

    /**
     * =============================END HOOKS==============================
     * @throws SmartyException
     */

    private function getRefundCommentBox()
    {
        $this->context->smarty->assign(
            [
                'verifone_refund_comment_box'    => $this->l('Refund Comment...'),
                'verifone_refund_max_length_str' => $this->l('*Max 150 characters.'),
                'verifone_refund_max_length'     => 150
            ]);

        return $this->context->smarty->fetch('module:verifone/views/templates/hook/refundCommentBox.tpl');
    }

    /**
     * @throws SmartyException
     */
    private function getTotalItemsForRefund($order)
    {
        $products    = $order->getProducts();
        $prods_array = [];
        foreach ($products as $prod_id => $params) {
            $prods_array[$prod_id] = $params['product_quantity'];
        }
        $notAllProductsAlert = $this->l('Please select all products!');
        $hasPaidShipping     = $this->hasPaidShipping($order->getShipping());
        $shippingMsgAlert    = $this->l('Order does not have free shipping. Please include that in full refund!');
        $this->context->smarty->assign(
            [
                'verifone_refund_products_list' => json_encode($prods_array),
                'not_all_products_alert'        => $notAllProductsAlert,
                'has_paid_shipping'             => $hasPaidShipping,
                'shipping_msg_alert'            => $shippingMsgAlert
            ]);

        return $this->context->smarty->fetch('module:verifone/views/templates/hook/refundItemsCorrection.tpl');
    }

    private function getAdminOrderPageRefundErrors()
    {
        $verifone_msg = '';
        if (isset($this->context->controller->errors['verifone_refund_error'])) {
            $verifone_msg = $this->context->controller->errors['verifone_refund_error'];
        }

        return $verifone_msg;
    }

    /**
     * @throws SmartyException
     */
    private function disablePartialRefund()
    {

        return $this->context->smarty->fetch('module:verifone/views/templates/hook/disablePartialRefund.tpl');
    }


    /**
     * @param $cart
     *
     * @return bool
     */
    public function checkCurrency($cart)
    : bool {
        $currency_order    = new Currency($cart->id_currency);
        $currencies_module = $this->getCurrency($cart->id_currency);

        if (is_array($currencies_module)) {
            foreach ($currencies_module as $currency_module) {
                if ($currency_order->id == $currency_module['id_currency']) {
                    return true;
                }
            }
        }

        return false;
    }

    public function getVerifonePaymentOption()
    : ?PaymentOption
    {

        $protocol = $this->isSSL() ? _PS_BASE_URL_SSL_ : _PS_BASE_URL_;
        $this->smarty->assign([
            'description' => $this->verifone_settings->getDescription(),
            'logo'        => $protocol.__PS_BASE_URI__.'modules/'.$this->name.'/logo.png'
        ]);

        $option = new PaymentOption();
        $option->setCallToActionText($this->verifone_settings->getTitle());

        if ($this->verifone_settings->getCheckoutType() === 'hpp') {
            $option->setAdditionalInformation($this->fetch('module:verifone/views/templates/hook/description.tpl'))
                   ->setAction('javascript:VerifoneCheckout()');
        } else {
            $option->setAdditionalInformation($this->fetch('module:verifone/views/templates/hook/iframe.tpl'))
                   ->setBinary(true);
        }

        return $option;
    }

    /**
     * generates the form for the payment option (hosted & inline)
     * @return string
     * @throws SmartyException
     */
    protected function generateForm()
    : string
    {
        return $this->context->smarty->fetch('module:verifone/views/templates/hook/payment_options.tpl');
    }

    /**
     * Resets the token on ANY setting change
     */
    public function reset_token()
    {
        if (false !== Configuration::get('__verifone_api_token')) {
            Configuration::deleteByName('__verifone_api_token');
        }
    }

    /**
     * @throws PrestaShopDatabaseException
     */
    public function checkActiveModule()
    : bool
    {
        $active  = false;
        $modules = Hook::getHookModuleExecList('paymentOptions') ?? [];
        foreach ($modules as $module) {
            if ($module['module'] == $this->name) {
                $active = true;
                break;
            }
        }

        return $active;
    }

    public function hasPaidShipping($shippingArr)
    : bool {
        $total = 0;
        foreach ($shippingArr as $valuesArr) {
            $total += $valuesArr['shipping_cost_tax_incl'];
        }

        return $total > 0;
    }


    /**
     * check if it has SSL
     * @return bool
     */
    private function isSSL()
    : bool
    {
        if (isset($_SERVER['HTTPS']) && $_SERVER['HTTPS'] === 'on') {
            return true;
        }

        if (isset($_SERVER['REQUEST_SCHEME']) && $_SERVER['REQUEST_SCHEME'] === 'https' ||
            isset($_SERVER['HTTP_X_FORWARDED_PROTO']) && $_SERVER['HTTP_X_FORWARDED_PROTO'] == 'https'
        ) {
            return true;
        }

        return false;
    }

    /**
     *  For debuting purposes
     *
     * @param $data
     * @param  bool  $stop
     *
     * @return void
     */
    function dump($data, bool $stop = true)
    {
        highlight_string("<?php \n".var_export($data, true).";\n?>");
        if ($stop) {
            exit;
        }
    }

    /**
     * @throws PrestaShopException
     * @throws PrestaShopDatabaseException
     */
    public function hookActionOrderStatusUpdate($params)
    {
        if ($params['newOrderStatus']->id == _PS_OS_REFUND_) {
            $order = new Order($params['id_order']);
            $this->verifoneRefund($order);
        }
    }

    private function verifoneRefund($order)
    {

        try {
            if (Validate::isLoadedObject($order) && $order->module === $this->name) {

                if ($order->getCurrentOrderState()->id != _PS_OS_REFUND_) {
                    $orderPayment   = OrderPayment::getByOrderReference($order->reference);
                    $transaction_id = null;
                    foreach ($orderPayment as $payment) {
                        $invoice = $payment->getOrderInvoice($order->id);
                        if ($invoice) {
                            $transaction_id = trim($payment->transaction_id);
                        }
                    }

                    //Prepare comment and get only the 150 accepted characters.
                    $refundCommentStr = Tools::getValue('vrefund-comment', '');
                    $refundComment    = trim(substr($refundCommentStr, 0, 150));

                    if ( ! $transaction_id) {
                        $this->logger->error($this->l('Cannot find a valid transaction ID for this order'));
                        $this->context->controller->errors[] = $this->l('Payment module encountered 
                    an error. Cannot get transaction id.');
                    }

                    $refund_class = new Verifone_Refund($order->total_paid, $refundComment, $transaction_id, $order);
                    $response     = $this->client->postRefund($refund_class->toArray(), $transaction_id);
                    if ( ! isset($response['id']) || ! isset($response['status']) || 'REFUNDED' !== $response['status']) {
                        $error_msg = sprintf($this->l('Error Refunding Invoice for order "%s" with transaction id "%s". Response: %s'),
                            $order->reference, $transaction_id, print_r($response, true));
                        $this->logger->error($error_msg);
                        $this->context->controller->errors[] = $error_msg;
                    } else {
                        $history           = new OrderHistory();
                        $history->id_order = (int) $this->order->id;
                        $history->changeIdOrderState(_PS_OS_REFUND_, $order, true);
                        $history->save();
                    }
                }
            }
        } catch (Exception $e) {
            $error_msg = sprintf($this->l('Exception, fatal error when trying to refund over Verifone API. Exception: %s'),
                $e->getMessage());
            $this->logger->error($error_msg);
            $this->context->controller->errors[] = $this->l('Fatal error when trying to do refund. 
            You need to initiate a full refund from Verifone portal.');
        }
    }
}
